//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/IO/IOFactory.h
//! @brief     Declares data import/export functions.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_IO_IOFACTORY_H
#define BORNAGAIN_DEVICE_IO_IOFACTORY_H

#include <string>

class Datafield;

namespace IO {

enum Filetype1D { any1D, bornagain1D, refltab, mft };
enum Filetype2D { any2D, bornagain, tiff, nicos };

#ifndef SWIG
bool isCompressed(const std::string& name);
Filetype1D filename2type1D(const std::string& filename);
Filetype2D filename2type2D(const std::string& filename);
#endif

//! Reads 1D data file and returns newly created Datafield object.
//! If selector is 'any1D', then the file extension will determine the loader to be used.
//! For any other selector, the file extension is ignored.
//! If file name ends with "*.gz" or "*.bz2" the file will be unzipped on the fly.
//! May throw, but will never return nullptr.
Datafield* readData1D(const std::string& file_name, Filetype1D ftype = any1D);

//! Reads 2D data file and returns newly created Datafield object.
//! If selector is 'any2D', then the file extension will determine the loader to be used.
//! For any other selector, the file extension is ignored.
//! If file name ends with "*.gz" or "*.bz2" the file will be unzipped on the fly.
//! May throw, but will never return nullptr.
Datafield* readData2D(const std::string& file_name, Filetype2D ftype = any2D);

//! Writes Datafield to file.
//! The format is determined by the file extension.
//! If file name ends with "*.gz" or "*.bz2" the file will be zipped on the fly.
void writeDatafield(const Datafield& data, const std::string& file_name);

} // namespace IO

namespace Util::RW {

//! Returns true if data in both files agree
bool filesAgree(const std::string& datFileName, const std::string& refFileName, double tol);

} // namespace Util::RW

#endif // BORNAGAIN_DEVICE_IO_IOFACTORY_H
