#ifndef PYOBJECTPTR_H
#define PYOBJECTPTR_H

#include "PyObjectDecl.h"
#include "Result.h"  // Status
#include <vector>

//========================================

// safe container for PyObjects
/*
The class `PyObjectPtr` contains a `PyObject*` (or `NULL`) and a `Status` which denotes the status of the value (OK/Warning/Error).
Decrementing Python reference-count is performed automatically when a
`PyObjectPtr` expires.
*/

class PyObjectPtr
{

public:
    // raw pointer to the PyObject
    PyObject* ptr = nullptr;

    // status of the PyObject
    Status status;

    PyObjectPtr(PyObject* pyobject_ptr, const Status& status);

    PyObjectPtr(PyObject* pyobject_ptr);

    PyObjectPtr(const Status& status);

    ~PyObjectPtr();

    // disallow copy constructor
    PyObjectPtr(const PyObjectPtr&) = delete;
    // disallow copy assignment
    PyObjectPtr& operator=(const PyObjectPtr&) = delete;
    // allow move constructor
    PyObjectPtr(PyObjectPtr&& other);
    // reset the container to the initial status (does _not_ release the Python resource)
    void reset();
    // reset the container to the initial status and return the PyObject pointer (does _not_ release the Python resource)
    PyObject* release();
    // discards the Python resource (decrements the Python reference)
    void discard();
    // check validity of the PyObjectPtr
    bool valid() const;
    // check if the PyObjectPtr is set
    bool isSet() const;
};


// Numpy array descriptor
struct NpArrayDescr
{
    // Numpy array descriptors
    bool C_contiguous = false;
    bool F_contiguous = false;
    // whether the data area of arr is aligned and in machine byte-order
    bool wellbehaved = false;
    // number of dimensions
    int n_dims = 0;
    // sizes of dimensions
    std::vector<np_size_t> dims;
    // character code indicating the data type
    char dtype = '\0';
    // whether Numpy array owns its data
    bool owns_data = false;
};

#endif  // PYOBJECTPTR_H
