/* Embedded Python Interpreter

Note that Python objects are structures allocated on the heap,
accessed through pointers of type `PyObject*`.

References:
- Python C-API <https://docs.python.org/3/c-api>
- Python ABI stability <https://docs.python.org/3/c-api/stable.html>
- Numpy C-API <https://numpy.org/doc/stable/reference/c-api>;
  <https://numpy.org/doc/stable/reference/c-api/array.html>.
- Python Extension Patterns <https://pythonextensionpatterns.readthedocs.io>
- "Python behind the scenes" series <https://tenthousandmeters.com/tag/python-behind-the-scenes>
- Python's garbage collector <https://rushter.com/blog/python-garbage-collector>
*/

#define INCLUDE_NUMPY  // also include Numpy via PyCore
#include "PyCore.h"
#undef INCLUDE_NUMPY

#include "PythonInterpreter.h"

#include <cstddef>  // NULL
#include <cstring>  // memcpy
#include <string>
#include <sstream>
#include <vector>
#include <memory>  // unique_ptr

#include <iostream>  // DEBUG
//========================================


namespace {
//-- Auxiliary functions

// thin wrapper to initialize Numpy
int _init_numpy()
{
    if(!PyArray_API) {
        /* To use Numpy Array C-API from an extension module,
           `import_array` function must be called.
           If the extension module is self-contained in a single .c file,
           then that is all that needs to be done.
           This function imports the module where the function-pointer table
           is stored and points the correct variable to it.
        */
        // NOTE: import_array returns `NULL` on failure.
        import_array();
        return 1; // Numpy Array API loaded successfully
    }

    return 2;  // Numpy Array API is already loaded
}


//! Converts char to string. In the case of nullptr will return an empty string
std::string toString(const char* const c)
{
    if (c)
        return c;
    return "";
}


std::string toString(const wchar_t* const c)
{
    if (!c)
        return "";
    std::wstring wstr(c);
    return std::string(wstr.begin(), wstr.end());
}

//! Converts PyObject into vector of strings, if possible, or throws exception
Result<std::vector<std::string>> toVectorString(PyObject* py_object)
{
    std::vector<std::string> result;

    if (PyTuple_Check(py_object)) {
        for (Py_ssize_t i = 0; i < PyTuple_Size(py_object); i++) {
            PyObject* value = PyTuple_GetItem(py_object, i);
            result.push_back(PythonInterpreter::pyStrtoString(value));
        }
    } else if (PyList_Check(py_object)) {
        for (Py_ssize_t i = 0; i < PyList_Size(py_object); i++) {
            PyObject* value = PyList_GetItem(py_object, i);
            result.push_back(PythonInterpreter::pyStrtoString(value));
        }
    } else
        return {Status(Status::Type::Error,
                       "Cannnot convert the given Python object "
                       "to vector<string>.")};
    return {std::move(result)};
}


// auxiliary function to convert a Numpy array data of real type `real_t`
// to vector<double>.
template<typename real_t>
inline void _realArray2DToDouble(PyArrayObject* const npArray_ptr,
                                 npy_intp rowIdxMax, npy_intp colIdxMax,
                                 std::vector<double>& vector_out)
{
    std::size_t idx = 0;
    for(npy_intp i_row = 0; i_row < rowIdxMax; ++i_row) {
        for(npy_intp j_col = 0; j_col < colIdxMax; ++j_col) {
            const double val = static_cast<double>(
                *reinterpret_cast<real_t*>(PyArray_GETPTR2(npArray_ptr, i_row, j_col)));
            vector_out[idx] = val;
            ++idx;
        }
    }
}


double* npyArray_to_CArray_WIP(PyObjectPtr& npy_array)
{
    if(!npy_array.valid() || !PyArray_Check(npy_array.ptr)) {
        throw std::runtime_error(
            "PythonInterpreter: Cannot convert an invalid Numpy array to a C-Array");
    }

    PyArrayObject* npArray_ptr {reinterpret_cast<PyArrayObject*>(npy_array.ptr)};
    PyArray_Descr* npArray_descr = PyArray_DESCR(npArray_ptr);
    const int npArray_type = PyArray_TYPE(npArray_ptr);
    const int npArray_ndim = PyArray_NDIM(npArray_ptr);
    npy_intp* npArray_dims = PyArray_DIMS(npArray_ptr);
    const int npArray_flags = PyArray_FLAGS(npArray_ptr);

    // DEBUG
    // type checking
    if(PyDataType_ISUNSIGNED(npArray_descr))
        printf("AN>> Array datatype = uint\n");
    else if(PyDataType_ISSIGNED(npArray_descr))
        printf("AN>> Array datatype = int\n");
    else if(PyDataType_ISFLOAT(npArray_descr))
        printf("AN>> Array datatype = float\n");
    else
        printf("AN>> Array datatype = ?\n");

    if(npArray_type == NPY_FLOAT)
        printf("AN>> Array type (%d) = f32\n", npArray_type);

    if(PyArray_IS_C_CONTIGUOUS(npArray_ptr))
        printf("AN>> Array is C-contiguous (row-major)\n");

    if(PyArray_IS_F_CONTIGUOUS(npArray_ptr))
        printf("AN>> Array is Fortran-contiguous (column-major)\n");

    if(PyArray_ISBEHAVED_RO(npArray_ptr) || PyArray_ISBEHAVED(npArray_ptr))
        printf("AN>> Array is well-behaved\n");

    if(npArray_flags & NPY_ARRAY_OWNDATA)
        printf("AN>> Array owns its data\n");
    else
        printf("AN>> Array does not own its data\n");

    for(int i_dim = 0; i_dim < npArray_ndim; ++i_dim)
        printf("AN>> Array-dim{%d} = %ld\n", i_dim, npArray_dims[i_dim]);

    // END DEBUG

    // TODO: Verify array is 2d

    // void *PyArray_GETPTR2(PyArrayObject *obj, npy_intp i, npy_intp j)
    // Quick, inline access to the element at the given coordinates in the ndarray, obj, which must have 2 dimensions (this is not checked). The corresponding i, j,  coordinates can be any integer but will be interpreted as npy_intp. You may want to typecast the returned pointer to the data type of the ndarray.

    // element-by-element copy of the Numpy array data (always work)
    const std::size_t vsize = static_cast<std::size_t>(PyArray_SIZE(npArray_ptr));
    std::vector<double> out1(vsize);
    for (std::size_t i = 0; i < vsize; i++) {
        out1[i] = *reinterpret_cast<double*>(PyArray_GETPTR1(npArray_ptr, i));
    }

    // using bulk memcpy-like operation, but keep in mind that NumPy ndarrays may be mis-aligned for the data type, have non-native byte order, or other subtle attributes that make such copies less than desirable.

    std::vector<double> out2(vsize);
    std::memcpy(out2.data(), PyArray_DATA(npArray_ptr),
                sizeof(double) * vsize);

    // using either approach, out now contains a copy of the ndarray's data

    double* npArray_data_ptr {reinterpret_cast<double*>(PyArray_DATA(npArray_ptr))};
    if(!npArray_data_ptr) {
        // checkError();
        throw std::runtime_error("PythonInterpreter: Cannot convert an invalid Numpy array to a C-Array (invalid data pointer)");
    }

    // TODO: Problem with the reference to numpy array!!!
    // returns a _new_ reference; ie. caller is responsible for the Python ref-count
    return npArray_data_ptr;
}


Result<double*> NpyArray1D_to_C(PyObject* pyobject_ptr)
{
    if(!pyobject_ptr || !PyArray_Check(pyobject_ptr)) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: Cannot convert an invalid "
                       "Numpy array to a C-Array")};
    }

    PyArrayObject* npArray_ptr {reinterpret_cast<PyArrayObject*>(pyobject_ptr)};
    PyArray_Descr* npArray_descr = PyArray_DTYPE(npArray_ptr);
    const int npArray_type = PyArray_TYPE(npArray_ptr);
    const int npArray_ndim = PyArray_NDIM(npArray_ptr);
    npy_intp* npArray_dims = PyArray_DIMS(npArray_ptr);
    const int npArray_flags = PyArray_FLAGS(npArray_ptr);
    // character code indicating the data type
    const char npArray_dtype = npArray_descr->type;

    //TODO: Enable type-casting for the Numpy array
    if(npArray_type != NPY_DOUBLE) {
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Expected a Numpy array of type _double_ "
               << "(given type '" << npArray_dtype << "')";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // DEBUG
    // type checking
    if(PyDataType_ISUNSIGNED(npArray_descr))
        printf("AN>> Array datatype = uint\n");
    else if(PyDataType_ISSIGNED(npArray_descr))
        printf("AN>> Array datatype = int\n");
    else if(PyDataType_ISFLOAT(npArray_descr))
        printf("AN>> Array datatype = float\n");
    else
        printf("AN>> Array datatype = ?\n");

    if(npArray_type == NPY_FLOAT)
        printf("AN>> Array type (%d) = f32\n", npArray_type);

    if(PyArray_IS_C_CONTIGUOUS(npArray_ptr))
        printf("AN>> Array is C-contiguous (row-major)\n");

    if(PyArray_IS_F_CONTIGUOUS(npArray_ptr))
        printf("AN>> Array is Fortran-contiguous (column-major)\n");

    if(PyArray_ISBEHAVED_RO(npArray_ptr) || PyArray_ISBEHAVED(npArray_ptr))
        printf("AN>> Array is well-behaved\n");

    if(npArray_flags & NPY_ARRAY_OWNDATA)
        printf("AN>> Array owns its data\n");
    else
        printf("AN>> Array does not own its data\n");

    for(int i_dim = 0; i_dim < npArray_ndim; ++i_dim)
        printf("AN>> Array-dim{%d} = %ld\n", i_dim, npArray_dims[i_dim]);

    // END DEBUG

    // TODO: check if array is contiguous
    double* npArray_data_ptr {reinterpret_cast<double*>(PyArray_DATA(npArray_ptr))};
    if(!npArray_data_ptr) {
        PythonInterpreter::checkError();
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot convert an invalid Numpy array to a C-Array"
                       "(data pointer is invalid)")};
    }

    return {std::move(npArray_data_ptr)};
}


PyObjectPtr call_pyfunction_npy_array(
    PyObjectPtr& py_module, const std::string& fn_name, PyObjectPtr& npy_array)
{
    // Call a Python function (from a given module) which takes a
    // single Numpy array as input and returns a Numpy array.

    if (!py_module.valid()) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter: Cannot call a function "
                       "from an invalid Python module")};
    }

    PyObject* pFunc = PyObject_GetAttrString(py_module.ptr, fn_name.c_str());
    if (!pFunc || !PyCallable_Check(pFunc)) {
        PythonInterpreter::checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter: The function '" << fn_name << "' "
               << "is not callable";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    PyObject* pReturn = PyObject_CallFunctionObjArgs(pFunc, npy_array.ptr,
                                                     NULL);
    if (!PyArray_Check(pReturn)) {
        PythonInterpreter::checkError();
        Py_DecRef(pFunc);
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter: Invalid return value from calling "
            "the function '" << fn_name << "' (expected a Numpy array)";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // returns a _new_ reference; ie. caller is responsible for the ref-count
    return {pReturn};
}

}  // namespace
//----------------------------------------

// Python stable ABI
Status PythonInterpreter::initialize()
{
    std::stringstream msg_ss;

    if (!Py_IsInitialized()) {
        Py_Initialize();
        msg_ss << "PythonInterpreter: Initialized Python - "
               << "version: " << Py_GetVersion();
        return {Status::Type::Info, msg_ss.str()};

    }

    // if Python already initialized
    msg_ss << "PythonInterpreter: Python already initialized - "
           << "version: " << Py_GetVersion();
    return {Status::Type::Warning, msg_ss.str()};
}

// Python stable ABI
bool PythonInterpreter::isInitialized()
{
    return static_cast<bool>(Py_IsInitialized());
}


// Python stable ABI
Status PythonInterpreter::finalize()
{
    // undo all initializations made by Py_Initialize() and subsequent use
    // of Python/C API functions, and destroy all sub-interpreters.
    // This is a no-op when called for a second time.
    Py_Finalize();

    std::stringstream msg_ss;
    msg_ss << "PythonInterpreter: Finalized Python - "
           << "version: " << Py_GetVersion();
    return {Status::Type::Info, msg_ss.str()};
}


// Python stable ABI
Status PythonInterpreter::checkError()
{
    if (PyErr_Occurred()) {
        // print a standard traceback to sys.stderr and clear the error indicator
        std::cerr << "---PythonInterpreter: Error in Python interpreter:\n";
        PyErr_Print();
        std::cerr << "\n---\n";
        return {Status::Type::Error};
    }

    return {Status::Type::Info, "No Python errors occured"};
}


// Python stable ABI
void PythonInterpreter::addPythonPath(const std::string& path)
{
    if (!path.empty()) {
        // add path to `PYTHONPATH`
        PyObject* sysPath = PySys_GetObject((char*)"path");
        PyList_Append(sysPath, PyUnicode_FromString(path.c_str()));
    }
}

// Python stable ABI
int PythonInterpreter::setPythonPath(const std::string& path)
{
    // returns 0 on success, -1 on error
    const int result = PySys_SetObject(
        (char*)"path", PyUnicode_FromString(path.c_str()));
    return result;
}

// Python stable ABI
PyObjectPtr PythonInterpreter::import(
    const std::string& pymodule_name, const std::string& path)
{
    addPythonPath(path);

    // import the module
    PyObject* pymodule = PyImport_ImportModule(pymodule_name.c_str());
    if (!pymodule || !PyModule_Check(pymodule)) {
        checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter: Cannot load Python module "
               << "'" << pymodule_name << "' "
               << "(given path = '" << path << "')";
        return PyObjectPtr(Status(Status::Type::Error, msg_ss.str()));
    }

    // returns a _new_ reference; ie. caller is responsible for the ref-count
    return {pymodule};
}


// Python stable ABI
void PythonInterpreter::DecRef(PyObject* py_object)
{
    Py_XDECREF(py_object);
}


std::string PythonInterpreter::pyStrtoString(PyObject* py_object)
{
    std::string result;
    PyObject* pyStr = PyUnicode_AsEncodedString(py_object, "utf-8", "replace");
    if (pyStr) {
        result = std::string(PyBytes_AsString(pyStr));
        Py_DecRef(pyStr);
    }

    return result;
}


std::string PythonInterpreter::runtimeInfo()
{
    std::stringstream result;

    // Runtime environment
    result << std::string(60, '=') << "\n";
    /* DEBUG

    result << "PATH: " << BaseUtils::System::getenv("PATH") << "\n";
    result << "PYTHONPATH: " << BaseUtils::System::getenv("PYTHONPATH") << "\n";
    result << "PYTHONHOME: " << BaseUtils::System::getenv("PYTHONHOME") << "\n";

    */

    // Embedded Python details
    result << "Py_GetProgramName(): " << toString(Py_GetProgramName()) << "\n";
    result << "Py_GetProgramFullPath(): " << toString(Py_GetProgramFullPath())
           << "\n";
    result << "Py_GetPath(): " << toString(Py_GetPath()) << "\n";
    result << "Py_GetPythonHome(): " << toString(Py_GetPythonHome()) << "\n";

    // Runtime Python's sys.path
    PyObject* sysPath = PySys_GetObject((char*)"path");
    Result<std::vector<std::string>> content {toVectorString(sysPath)};
    if(content.valid()) {
        result << "sys.path: ";
        for (auto s : content.val)
            result << s << ",";
        result << "\n";
    }

    return result.str();
}

// Attempt to retrieve Python stack trace
// Ref: <https://stackoverflow.com/q/1796510>
std::string PythonInterpreter::stackTrace()
{
    std::stringstream result;

    if (PyErr_Occurred()) {
        PyObject *ptype, *pvalue, *ptraceback, *pystr;

        PyErr_Fetch(&ptype, &pvalue, &ptraceback);
        pystr = PyObject_Str(pvalue);
        if (char* str = PyBytes_AsString(pystr))
            result << std::string(str) << "\n";
        Py_DecRef(pystr);

        PyObject* module_name = PyUnicode_FromString("traceback");
        PyObject* py_traceback_module = PyImport_Import(module_name);
        Py_DecRef(module_name);

        if (py_traceback_module) {
            result << "\n";
            PyObject* pyth_func = PyObject_GetAttrString(py_traceback_module, "format_exception");
            if (pyth_func && PyCallable_Check(pyth_func)) {
                PyObject* pyth_val =
                    PyObject_CallFunctionObjArgs(pyth_func, ptype, pvalue, ptraceback, NULL);
                Py_DecRef(pyth_func);
                if (pyth_val) {
                    pystr = PyObject_Str(pyth_val);
                    if (char* str = PyBytes_AsString(pystr))
                        result << std::string(str);
                    Py_DecRef(pystr);
                    Py_DecRef(pyth_val);
                }
            }
            result << "\n";
        }

        Py_DecRef(py_traceback_module);
    }

    result << "\n";
    result << PythonInterpreter::runtimeInfo();
    result << "\n";

    return result.str();
}


std::string PythonInterpreter::errorDescription(const std::string& title)
{
    std::stringstream msg;
    msg << title << "\n";
    msg << PythonInterpreter::stackTrace() << "\n";
    return msg.str();
}

//----------------------------------------

Status PythonInterpreter::Numpy::initialize() {

    // initialize Python C API, if needed
    Status py_status{PythonInterpreter::initialize()};
    if(py_status.isError())
        return py_status;

    const int res {_init_numpy()};

    switch(res)
    {
    case 1:
        return {Status::Type::Info, "PythonInterpreter: Initialized Numpy"};
    case 2:
        return {Status::Type::Warning, "PythonInterpreter: Numpy already initialized"};
    default:
        return {Status::Type::Error, "PythonInterpreter: Cannot initialize Numpy"};
    }
}


bool PythonInterpreter::Numpy::isInitialized()
{
    return static_cast<bool>(PyArray_API);
}

//----------------------------------------

Result<NpArrayDescr> PythonInterpreter::Numpy::arrayDescriptor(
    PyObject* pyobject_ptr)
{
    if(!pyobject_ptr || !PyArray_Check(pyobject_ptr)) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Invalid Numpy array")};
    }

    PyArrayObject* npArray_ptr {reinterpret_cast<PyArrayObject*>(pyobject_ptr)};
    PyArray_Descr* npArray_descr = PyArray_DTYPE(npArray_ptr);
    // const int npArray_type = PyArray_TYPE(npArray_ptr);
    const int npArray_ndim = PyArray_NDIM(npArray_ptr);
    npy_intp* npArray_dims = PyArray_DIMS(npArray_ptr);
    const int npArray_flags = PyArray_FLAGS(npArray_ptr);
    // character code indicating the data type
    const char npArray_dtype = npArray_descr->type;


    NpArrayDescr np_descr_out;
    np_descr_out.n_dims = npArray_ndim;
    if (np_descr_out.n_dims > 0) {
        np_descr_out.dims.resize(npArray_ndim);
        for(npy_intp i_d = 0; i_d < npArray_ndim; ++i_d)
            np_descr_out.dims[i_d] = npArray_dims[i_d];
    }

    np_descr_out.dtype = npArray_dtype;

    if(PyArray_IS_C_CONTIGUOUS(npArray_ptr))
        np_descr_out.C_contiguous = true;

    if(PyArray_IS_F_CONTIGUOUS(npArray_ptr))
        np_descr_out.F_contiguous = true;

    if(PyArray_ISBEHAVED_RO(npArray_ptr))
        np_descr_out.wellbehaved = true;

    if(npArray_flags & NPY_ARRAY_OWNDATA)
        np_descr_out.owns_data = true;

    return {std::move(np_descr_out)};
}


Result<std::vector<double>> PythonInterpreter::Numpy::createVectorFromArray2D(PyObject* pyobject_ptr)
{
    if(!pyobject_ptr || !PyArray_Check(pyobject_ptr)) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: Cannot convert an invalid "
                       "Numpy array to a C-Array")};
    }

    PyArrayObject* npArray_ptr {reinterpret_cast<PyArrayObject*>(pyobject_ptr)};
    PyArray_Descr* npArray_descr = PyArray_DTYPE(npArray_ptr);
    const int npArray_type = PyArray_TYPE(npArray_ptr);
    const int npArray_ndim = PyArray_NDIM(npArray_ptr);
    npy_intp* npArray_dims = PyArray_DIMS(npArray_ptr);
    const std::size_t npArray_size = static_cast<std::size_t>
        (PyArray_SIZE(npArray_ptr));
    const int npArray_flags = PyArray_FLAGS(npArray_ptr);
    // character code indicating the data type
    const char npArray_dtype = npArray_descr->type;

    // Numpy array type must be 2d
    if(npArray_ndim != 2) {
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Expected a Numpy 2d-array "
               << "(given number of dimensions is " << npArray_ndim << ")";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // Numpy array type must be numeric and real (eligible for casting to double)
    if(!PyDataType_ISNUMBER(npArray_descr) || PyDataType_ISCOMPLEX(npArray_descr)) {
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Expected a Numpy array of numeric type and real "
               << "(given type '" << npArray_dtype << "')";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    std::vector<double> data(npArray_size);

    const npy_intp rowIdxMax = npArray_dims[0], colIdxMax = npArray_dims[1];

    // branch on type
    switch(npArray_type)
    {
    // float
    case NPY_DOUBLE:
        _realArray2DToDouble<npy_double>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
    case NPY_FLOAT:
        _realArray2DToDouble<npy_float>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
    // signed int
    case NPY_INT8:
        _realArray2DToDouble<npy_int8>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
    case NPY_INT16:
        _realArray2DToDouble<npy_int16>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
    case NPY_INT32:
        _realArray2DToDouble<npy_int32>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;

// prevent the error from Windows compiler MSVC 19:
// `error C2196: case value 'NPY_LONG' already used`
#ifndef _WIN32

    case NPY_LONG:
        _realArray2DToDouble<npy_long>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
#endif // _WIN32

    // unsigned int
    case NPY_UINT8:
        _realArray2DToDouble<npy_uint8>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
    case NPY_UINT16:
        _realArray2DToDouble<npy_uint16>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
    case NPY_UINT32:
        _realArray2DToDouble<npy_uint32>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;

// prevent the error from Windows compiler MSVC 19:
// `error C2196: case value 'NPY_ULONG' already used`
#ifndef _WIN32
    case NPY_ULONG:
        _realArray2DToDouble<npy_ulong>(npArray_ptr, rowIdxMax, colIdxMax, data);
        break;
#endif // _WIN32

    default:
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Conversion of Numpy array of dtype '" << npArray_dtype << "' "
               << "to vector<double> is not implemented";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    std::stringstream msg_ss;
    msg_ss << "PythonInterpreter::Numpy: "
           << "Numpy 2d-array of dtype '" << npArray_dtype << "' "
           << "and shape (" << npArray_dims[0] << "," << npArray_dims[1] << ") "
           << "converted to vector<double> with size " << data.size();

    return {std::move(data), Status(Status::Type::Info, msg_ss.str())};
}


PyObjectPtr PythonInterpreter::Numpy::createArray2DfromC(
    double* const c_array, const np_size_t dims[2])
{
    if (!c_array) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot create a Numpy 1D-array from a null data pointer")};
    }

    const np_size_t size = dims[0] * dims[1];
    if (size < 1) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot create a Numpy 1D-array from a data with size = 0")};
    }

    npy_intp npDims[2] = {dims[0], dims[1]};

    // create stand-alone Numpy array (float64)
    PyObject* npArray_ptr = PyArray_SimpleNew(/* n_dims */ 2,
                                              npDims, NPY_DOUBLE);
    if (!npArray_ptr) {
        checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Cannot create a Numpy 1D-array from the "
               << "given data (size = " << size << ")";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // obtain pointer to data buffer of Numpy array
    double* array_buffer = static_cast<double*>(
        PyArray_DATA((PyArrayObject*)(npArray_ptr)));

    for (np_size_t index = 0; index < size; ++index)
    {
        *array_buffer = c_array[index];
        ++array_buffer;
    }

    // returns a _new_ reference; ie. caller is responsible for the ref-count
    return {npArray_ptr};
}


PyObjectPtr PythonInterpreter::Numpy::createArray1DfromC(
    double* const c_array, const np_size_t size)
{
    if (!c_array) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot create a Numpy 1D-array from a null data pointer")};
    }

    if (size < 1) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot create a Numpy 1D-array from a data with size = 0")};
    }

    npy_intp npDims[1] = {size};

    // create stand-alone Numpy array (float64)
    PyObject* npArray_ptr = PyArray_SimpleNew(/* n_dims */ 1,
                                              npDims, NPY_DOUBLE);
    if (!npArray_ptr) {
        checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Cannot create a Numpy 1D-array from the "
               << "given data (size = " << size << ")";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // obtain pointer to data buffer of Numpy array
    double* array_buffer = static_cast<double*>(
        PyArray_DATA((PyArrayObject*)(npArray_ptr)));

    for (np_size_t index = 0; index < size; ++index)
    {
        *array_buffer = c_array[index];
        ++array_buffer;
    }

    // returns a _new_ reference; ie. caller is responsible for the ref-count
    return {npArray_ptr};
}


PyObjectPtr PythonInterpreter::Numpy::CArrayAsNpy2D(
    double* const c_array, const np_size_t dims[2])
{
    if (!c_array) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot create a Numpy 2D-array from a null data pointer")};
    }

    const np_size_t size = dims[0] * dims[1];
    if (size < 1) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: "
                       "Cannot create a Numpy 2D-array from a data with size = 0")};
    }

    npy_intp npDims[2] = {dims[0], dims[1]};

    // convert the contiguous C-array to Numpy array
    PyObject* npArray_ptr = PyArray_SimpleNewFromData(
        /* n_dims */ 2, npDims, NPY_DOUBLE, reinterpret_cast<void*>(c_array));

    if(!npArray_ptr || !PyArray_Check(npArray_ptr)) {
        PythonInterpreter::checkError();
        return {Status(Status::Type::Error,
                       "PythonInterpreter::Numpy: Cannot convert the given "
                       "C-Array to a Numpy 2D-array")};
    }

    // returns a _new_ reference; ie. caller is responsible for the ref-count
    return {npArray_ptr};
}


PyObjectPtr PythonInterpreter::Numpy::arrayND(std::vector<std::size_t>& dimensions)
{
    // descriptors for the array dimensions
    const std::size_t n_dims = dimensions.size();
    if (n_dims < 1) {
        return {Status(Status::Type::Error,
                       "Cannot make a Numpy with the given number of dimensions; "
                       "number of dimensions must be >= 1")};
    }

    for (std::size_t d = 0; d < n_dims; ++d) {
        if (dimensions[d] < 2) {
            return {Status(Status::Type::Error,
                           "Cannot make a Numpy with the given dimensions; "
                           "dimensions must be >= 2")};
        }
    }

    npy_int ndim_numpy = static_cast<npy_int>(n_dims);
    npy_intp* ndimsizes_numpy = new npy_intp[n_dims];
    for (std::size_t d = 0; d < n_dims; ++d)
        ndimsizes_numpy[d] = dimensions[d];

    // creating standalone numpy array
    PyObject* npyArray_ptr = PyArray_SimpleNew(
        ndim_numpy, ndimsizes_numpy, NPY_DOUBLE);
    delete[] ndimsizes_numpy;

    if (!npyArray_ptr) {
        checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Cannot create a Numpy " << n_dims << "D-array "
               << "from the given data";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    return {npyArray_ptr};
}


Result<double*> PythonInterpreter::Numpy::getDataPtr(PyObject* pyobject_ptr)
{

    PyArrayObject* npArray_ptr {reinterpret_cast<PyArrayObject*>(pyobject_ptr)};

    // get the pointer to data buffer of the Numpy array
    double* data_ptr = static_cast<double*>(
        PyArray_DATA((PyArrayObject*)npArray_ptr));

    if (!data_ptr) {
        checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter::Numpy: "
               << "Numpy array has invalid data pointer";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    return {std::move(data_ptr)};
}

//----------------------------------------

PyObjectPtr PythonInterpreter::BornAgain::import(const std::string& path)
{
    PythonInterpreter::addPythonPath(path);

#ifndef _WIN32
    // Stores signal handler before numpy's mess it up.
    // This is to make ctrl-c working from terminal.

    PyOS_sighandler_t sighandler;
    sighandler = PyOS_getsig(SIGINT);
#endif

    PyObject* ba_pymodule = PyImport_ImportModule("bornagain");

    if (!ba_pymodule || !PyModule_Check(ba_pymodule)) {
        checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter: Cannot load 'bornagain' Python module "
               << "(given path = '" << path << "')";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // restores single handler to make ctr-c alive.
#ifndef _WIN32
    PyOS_setsig(SIGINT, sighandler);
#endif

    return {ba_pymodule};
}


PyObjectPtr PythonInterpreter::BornAgain::importScript(
    const std::string& script, const std::string& path)
{
    PyObjectPtr ba_pymodule {PythonInterpreter::BornAgain::import(path)};
    // TODO: Check ba module

    PyObject* pCompiledFn = Py_CompileString(script.c_str(), "", Py_file_input);
    if (!pCompiledFn) {
        ba_pymodule.discard();
        return {Status(Status::Type::Error,
                       PythonInterpreter::errorDescription(
                           "Cannot compile Python snippet"))};
    }

    // create a module
    PyObject* tmpModule = PyImport_ExecCodeModule((char*)"tmpModule", pCompiledFn);
    if (!tmpModule) {
        Py_DecRef(pCompiledFn);
        ba_pymodule.discard();
        return {Status(Status::Type::Error,
                       PythonInterpreter::errorDescription(
                           "Cannot execute Python snippet"))};
    }

    return {tmpModule};
}


Result<std::vector<std::string>> PythonInterpreter::BornAgain::listOfFunctions(
    const std::string& script, const std::string& path)
{
    PyObjectPtr tmpModule {PythonInterpreter::BornAgain::importScript(script, path)};

    PyObject* pDict = PyModule_GetDict(tmpModule.ptr);
    if (!pDict) {
        return {Status(Status::Type::Error, "PythonInterpreter::BornAgain: "
                       "Cannot obtain the dictionary from the script module")};
    }

    PyObject *key, *value;
    Py_ssize_t pos = 0;
    std::vector<std::string> fn_names;
    while (PyDict_Next(pDict, &pos, &key, &value)) {
        if (PyCallable_Check(value)) {
            std::string func_name {PythonInterpreter::pyStrtoString(key)};
            if (func_name.find("__") == std::string::npos)
                fn_names.push_back(func_name);
        }
    }

    Py_DecRef(pDict);

    return {std::move(fn_names)};
}

//----------------------------------------

PyObjectPtr PythonInterpreter::Fabio::import()
{
    return {PythonInterpreter::import("fabio")};
}


PyObjectPtr PythonInterpreter::Fabio::open(const std::string& filename,
                                           PyObjectPtr& fabio_module)
{
    // load an image via calling `fabio.load` function which takes a
    // filename (Python string) and returns a Numpy array.

    if (!fabio_module.valid() || !PyModule_Check(fabio_module.ptr)) {
        return {Status(Status::Type::Error,
                       "PythonInterpreter.fabio: Invalid Python module "
                       "(expected 'fabio' module)")};
    }

    PyObject* pFunc = PyObject_GetAttrString(fabio_module.ptr, (char*)"open");
    if (!pFunc || !PyCallable_Check(pFunc)) {
        PythonInterpreter::checkError();
        return {Status(Status::Type::Error,
                       "PythonInterpreter.fabio: The function 'fabio.open' is not callable")};
    }

    // convert the filename to a Python unicode string
    PyObject* pFilename = PyUnicode_FromString(filename.c_str());
    if (!pFilename) {
        PythonInterpreter::checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter.fabio: Filename '" << filename << "' "
               << "cannot be converted to Python unicode string";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // get the result of the call `fabio.open(<filename>)`
    PyObject* pResult_open = PyObject_CallFunctionObjArgs(pFunc, pFilename, NULL);
    Py_DecRef(pFunc);
    if (!pResult_open) {
        PythonInterpreter::checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter.fabio: Invalid return value from "
               << "calling the function 'fabio.open(\"" << filename << "\")'";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // get `result.data` (must be a Numpy array)
    PyObject* npyArray_ptr = PyObject_GetAttrString(pResult_open, (char*)"data");
    Py_DecRef(pResult_open);
    if (!npyArray_ptr || !PyArray_Check(npyArray_ptr)) {
        PythonInterpreter::checkError();
        std::stringstream msg_ss;
        msg_ss << "PythonInterpreter.fabio: Invalid return value from "
               << "calling the function 'fabio.open(\"" << filename << "\")' "
               << "(expected a Numpy array)";
        return {Status(Status::Type::Error, msg_ss.str())};
    }

    // returns a _new_ reference; ie. caller is responsible for the ref-count
    return {npyArray_ptr};
}
