#ifndef PYTHONINTRP_H
#define PYTHONINTRP_H

#include <vector>
#include <string>
#include "PyObjectPtr.h"

class MultiLayer;

// Embedded Python interpreter
namespace PythonInterpreter {

// Python stable ABI
Status initialize();

// Python stable ABI
bool isInitialized();

// Python stable ABI
Status finalize();

// Python stable ABI
Status checkError();

// Python stable ABI
void addPythonPath(const std::string& path);

// Python stable ABI
int setPythonPath(const std::string& path);

// Python stable ABI
PyObjectPtr import(const std::string& pymodule_name,
                   const std::string& path = "");
// Python stable ABI
void DecRef(PyObject* py_object);

//! Converts String PyObject into string, if possible, or throws exception
std::string pyStrtoString(PyObject* obj);

//! Returns multi-line string representing PATH, PYTHONPATH, sys.path and other info.
std::string runtimeInfo();

//! Returns string representing python stack trace.
std::string stackTrace();

//! Returns the full error description via Python stack trace
std::string errorDescription(const std::string& title = "");

//-- Numpy-related functionality
namespace Numpy {

// initialize Numpy
Status initialize();

// Python stable ABI
bool isInitialized();

//! Obtains the descriptor of a Numpy array
Result<NpArrayDescr> arrayDescriptor(PyObject* pyobject_ptr);

//! Creates an empty Numpy array of type `double` with the given dimensions
PyObjectPtr arrayND(std::vector<std::size_t>& dimensions);

//! Returns the pointer to the data buffer of a Numpy array of type `double`
Result<double*> getDataPtr(PyObject* pyobject_ptr);

//! Creates a Numpy 1D-array from a contiguous C-array of type `double`.
// Data will be copied.
PyObjectPtr createArray1DfromC(double* const c_array, const np_size_t size);

// !Creates a Numpy 2D-array from a contiguous C-array of type `double`.
// Data will be copied.
PyObjectPtr createArray2DfromC(double* const c_array, const np_size_t dims[2]);

// !Creates a `vector<double>` from a Numpy 2D-array (data will be copied).
Result<std::vector<double>> createVectorFromArray2D(PyObject* pyobject_ptr);

// !Creates a 2D Numpy array from a contiguous C-array of type `double`.
// The data is _not_ copied, and the Numpy array will _not_ own its data.
PyObjectPtr CArrayAsNpy2D(double* const c_array, const np_size_t dims[2]);

}  // namespace Numpy

//-- BornAgain-related functionality
namespace BornAgain {
//! Imports BornAgain from given location. If path is empty, tries to rely on PYTHONPATH.
PyObjectPtr import(const std::string& path = "");

// import a 'BornAgain' Python script
PyObjectPtr importScript(const std::string& script,
                         const std::string& path);

//! Returns list of functions defined in the script.
//! @param script: Python script
//! @param path: A path to import BornAgain library. If empty, relies on PYTHONPATH
Result<std::vector<std::string>> listOfFunctions(
    const std::string& script, const std::string& path);

}  // namespace BornAgain

//-- Fabio-related functionality
namespace Fabio {
// import 'fabio' Python module
PyObjectPtr import();
// call 'fabio.open' on a given file (needs Numpy)
PyObjectPtr open(const std::string& filename,
                 PyObjectPtr& fabio_module);

} // namespace Fabio

}  // namespace PythonInterpreter

#endif  // PYTHONINTRP_H
