#ifndef RESULT_H
#define RESULT_H

#include <utility>  // move
#include <string>

// ========================================

// Indicator for the status of a result
struct Status
{
    enum class Type {None=-1, Info, Warning, Error};
    Type type = Type::None;

    // status message
    std::string message;

    Status(const Status::Type type = Type::None,
           const std::string& message = "");

    bool isSet() const;

    bool isOK() const;

    bool isWarning() const;

    bool isError() const;

    // cast to boolean (true if result is valid)
    operator bool() const;
};

//========================================

// Container for Results which may fail
/* A `Result<T>` instance contains a value of type `T` (or none) and
   a `Status` which denotes the status of the value (OK/Warning/Error).
   This is used to avoid throwing exceptions in a _core_ module,
   leaving the decision to the callers (or the higher interface) whether
   an exception should be raised or not.
*/
template <typename ResType>
class Result
{

public:

    using type = ResType;

    // stored result
    ResType val;

    // status of the PyObject
    Status status;

    Result(ResType&& result, Status&& status);

    Result(ResType&& result);

    Result(Status&& status);

    ~Result();

    // disallow copy constructor
    Result(const Result&) = delete;
    // disallow copy assignment
    Result& operator=(const Result&) = delete;
    // allow move constructor
    Result(Result&& other);
    // reset the container to the initial status
    void reset();
    // check validity of the Result
    bool valid() const;
    // check if the Result is set
    bool isSet() const;
};


//-- Implementation of Result

template <typename ResType>
Result<ResType>::Result(ResType&& result,
                        Status&& status):
    val{std::move(result)},
    status{std::move(status)} {}


template <typename ResType>
Result<ResType>::Result(ResType&& result):
    Result(std::move(result), Status(Status::Type::Info)) {}


template <typename ResType>
Result<ResType>::Result(Status&& status):
    status{std::move(status)} {}


template <typename ResType>
Result<ResType>::~Result() {}


template <typename ResType>
Result<ResType>::Result(Result<ResType>&& other):
    Result(std::move(other.val), std::move(other.status))
{
    // reset the moved object
    other.reset();
}


template <typename ResType>
void Result<ResType>::reset()
{
    status = Status();
}


template <typename ResType>
bool Result<ResType>::valid() const
{
    return (status.isOK() || status.isWarning());
}


template <typename ResType>
bool Result<ResType>::isSet() const
{
    return (status.isSet());
}

#endif  // RESULT_H
